import warnings

warnings.warn("pymatgen.serializers.json_coders has been moved to "
              "pymatgen.util.serialization. This stub will be removed in pmg "
              "2018.01.01.")
from pymatgen.util.serialization import *

