import warnings

warnings.warn("pymatgen.phasediagram and submodules has been moved to "
              "pymatgen.analysis.phase_diagram This stub will be "
              "removed in pmg 2018.01.01.")

from pymatgen.analysis.phase_diagram import *

